from django.urls import path
from goflow.graphics2 import views as goflow_graphics2_views


urlpatterns = [
    path('processimage/<int:process_id>/pos_activity/', goflow_graphics2_views.pos_activity),
]



#!/usr/local/bin/python
# -*- coding: utf-8 -*-


from django.forms import ModelForm
from django import forms
from datetime import datetime

from django.contrib.contenttypes.models import ContentType

from goflow.apptools.models import DefaultAppModel
from django.utils.translation import gettext_lazy as _, gettext


class BaseForm(ModelForm):
    '''
    base class for edition forms
    '''
    
    workitem_id = forms.IntegerField(widget=forms.HiddenInput, required=False)
    priority = forms.ChoiceField(label=_(u'Priority'), widget=forms.RadioSelect, initial='0',
                                 choices=(('0',_('normal')), ('1',_('urgent')), ('5',_('emergency'))),
                                 )
    def save(self, workitem=None, submit_value=None, commit=True):
        ob = super(BaseForm, self).save(commit=commit)
        if workitem and workitem.can_priority_change():
            workitem.priority = int(self.cleaned_data['priority'])
            workitem.save()
        return ob
    
    def pre_check(self, obj_context=None, user=None):
        """may be overriden to do some check before.
        
        obj_context    object instance (if cmp_attr is set, this is the root object)
        an exception should be risen if pre-conditions are not fullfilled
        """
        pass

    class Meta:
         exclude = ('workitem_id',)

class StartForm(ModelForm):
    '''
    base class for starting a workflow
    '''
    priority = forms.ChoiceField(label=_('Priority'), widget=forms.RadioSelect, initial='0',
                                 choices=(('0','normal'), ('1','urgent'), ('5','emergency')),
                                 )
    
    def save(self, user=None, data=None, commit=True):
        ob = super(StartForm, self).save(commit=commit)
        return ob
    
    def pre_check(self, user=None):
        """may be overriden to do some check before.
        
        an exception should be risen if pre-conditions are not fullfilled
        """
        pass


class DefaultAppForm(BaseForm):
    def save(self, workitem=None, submit_value=None, commit=True):
        ob = super(DefaultAppForm, self).save(commit=False)
        if ob.comment:
            if not ob.history:
                ob.history = gettext('Init')
            ob.history += '\n---------'
            if workitem:
                ob.history += gettext('\nActivity: [%s]') % workitem.activity.title
            ob.history += '\n%s\n%s' % (datetime.now().isoformat(' '), ob.comment)
            ob.comment = None
        if submit_value:
            if ob.history:
                ob.history += gettext('\n button clicked: [%s]') % submit_value
        ob.save()
        return ob

    class Meta:
         model = DefaultAppModel
         exclude = ('reasonDenial',)


class DefaultAppStartForm(StartForm):
    def save(self,  user=None, data=None, commit=True):
        ob = super(DefaultAppStartForm, self).save(commit=False)
        if not ob.history:
            ob.history = gettext('Init')
        ob.history += gettext('\n%s start instance') % datetime.now().isoformat(' ')
        if ob.comment:
            ob.history += '\n---------'
            ob.history += '\n%s' % ob.comment
            ob.comment = None
        ob.save()
        return ob

    class Meta:
         model = DefaultAppModel
         exclude = ('reasonDenial',)


ctypes = ContentType.objects.\
        exclude(app_label='auth').\
        exclude(app_label='contenttypes').\
        exclude(app_label='workflow').\
        exclude(app_label='graphics').\
        exclude(app_label='graphics2').\
        exclude(app_label='runtime').\
        exclude(app_label='apptools').\
        exclude(app_label='sessions').\
        exclude(app_label='sites').\
        exclude(app_label='admin')

class ContentTypeForm(forms.Form):
    ctype = forms.ModelChoiceField(
                queryset=ctypes, 
                required=True, 
                empty_label=_('(select a content-type)'),
                label=_('content type'),
                help_text=_('clone all instances of the selected content type and push '
                           'them in the test process of the application')
            )



#!/usr/local/bin/python
# -*- coding: utf-8 -*-
import json

from django.shortcuts import render
from django.conf import settings
from django.contrib.auth.models import Group
from django.contrib.auth import authenticate, login, logout
from django.contrib.contenttypes.models import ContentType
from django.http import HttpResponseRedirect, HttpResponse, JsonResponse, HttpResponseForbidden, HttpResponseBadRequest

from goflow.runtime.models import WorkItem
from goflow.workflow.models import *
from django.utils.translation import gettext_lazy as _, gettext
from django.contrib import messages


def index(request, template='workflow/index.html', extra_context={}):
    """workflow dashboard handler.
    
    template context contains following objects:
    - user
    - processes
    - roles
    
    other applications (ie runtime or apptools) should fill extra_context.
    """
    print('index request\n\n\n')
    me = request.user
    roles = Group.objects.all()
    processes = Process.objects.all()
    # optional package (ugly design)
    try:
        from goflow.apptools.models import DefaultAppModel
        obinstances = DefaultAppModel.objects.all()
    except Exception:
        obinstances = None
    
    context = {'processes':processes, 'roles':roles}
    context.update(extra_context)
    return render(request, template, context)


def debug_switch_user(request, username, password, redirect=None):
    """
    fast user switch for test purpose.
    
    parameters:
    
    username
        username
    password
        password
    redirect
        redirection url
    
    *FOR TEST ONLY*
    
    see template tag switch_users.
    """
    logout(request)
    # return HttpResponseRedirect(redirect)
    if not redirect:
        redirect = request.META['HTTP_REFERER']
    user = authenticate(username=username, password=password)
    if user is not None:
        if user.is_active:
            login(request, user)
            return HttpResponseRedirect(redirect)
        else:
            return HttpResponse(gettext('user is not active'))
    else:
        return HttpResponse(gettext('authentication failed'))


def userlist(request, template):
    '''
    not used
    '''
    return HttpResponse(gettext('user page.'))


def process_dot(request, id, template='goflow/process.dot'):
    """graphviz generator (**Work In Progress**).
    (**Work In Progress**)
    
    id process id
    template graphviz template
    
    context provides: process, roles, activities
    """
    process = Process.objects.get(id=int(id))
    context = {
               'process': process,
               'roles': ({'name':'role1', 'color':'red'},),
               'activities': Activity.objects.filter(process=process)
               }
    return render(request, template, context)


def cron(request=None):
    """
    (**Work In Progress**)
    TODO: move to instances ?
    """
    for t in Transition.objects.filter(condition__contains='workitem.timeout'):
        workitems = WorkItem.objects.filter(
            activity=t.input).exclude(status='complete')
        for wi in workitems:
            wi.forward(timeout_forwarding=True)
    
    if request:
        #request.user.message_set.create(message="cron has run.")
        messages.add_message(request, messages.INFO, "cron has run.")
        #if request.META.has_key('HTTP_REFERER'):
        if 'HTTP_REFERER' in request.META.keys():
            url = request.META['HTTP_REFERER']
        else:
            url = 'home/'
        return HttpResponseRedirect(url)


def workflow_designer(request, process_id):
    if not request.user.is_authenticated or not request.user.is_staff:
        return HttpResponseForbidden("Forbidden")
    process = Process.objects.get(id=process_id)
    return render(request, "goflow/workflow_designer.html", {"process": process})


def workflow_graph(request, process_id):
    if not request.user.is_authenticated or not request.user.is_staff:
        return HttpResponseForbidden("Forbidden")
    process = Process.objects.get(id=process_id)

    layout = WorkflowLayout.objects.filter(process=process).first()
    layout_map = layout.layout if layout else {}

    if request.method == "GET":
        activities = []
        for activity in Activity.objects.filter(process=process):
            position = layout_map.get(str(activity.id))
            activities.append(
                {
                    "id": activity.id,
                    "title": activity.title,
                    "kind": activity.kind,
                    "autostart": activity.autostart,
                    "autofinish": activity.autofinish,
                    "position": position,
                }
            )
        transitions = Transition.objects.filter(process=process).values(
            "id",
            "name",
            "input_id",
            "output_id",
            "condition",
        )
        payload = {
            "process": {
                "id": process.id,
                "title": process.title,
                "begin_id": process.begin_id,
                "end_id": process.end_id,
            },
            "activities": list(activities),
            "transitions": list(transitions),
        }
        return JsonResponse(payload)

    if request.method != "POST":
        return HttpResponseBadRequest("Unsupported method")

    try:
        data = json.loads(request.body.decode("utf-8"))
    except Exception:
        return HttpResponseBadRequest("Invalid JSON")

    activities = data.get("activities", [])
    transitions = data.get("transitions", [])
    positions = data.get("positions", {})
    deleted_activity_ids = set(data.get("deleted_activity_ids", []))
    deleted_transition_ids = set(data.get("deleted_transition_ids", []))
    begin_id = data.get("begin_id")
    end_id = data.get("end_id")

    for activity_id in deleted_activity_ids:
        Activity.objects.filter(process=process, id=activity_id).delete()
        layout_map.pop(str(activity_id), None)

    for transition_id in deleted_transition_ids:
        Transition.objects.filter(process=process, id=transition_id).delete()

    activity_map = {}
    for item in activities:
        item_id = item.get("id")
        defaults = {
            "title": item.get("title", "Activity"),
            "kind": item.get("kind", "standard"),
            "autostart": bool(item.get("autostart", False)),
            "autofinish": bool(item.get("autofinish", True)),
        }
        if item_id:
            activity = Activity.objects.get(id=item_id, process=process)
            for field, value in defaults.items():
                setattr(activity, field, value)
            activity.save()
        else:
            activity = Activity.objects.create(process=process, **defaults)
        activity_map[item.get("client_id") or activity.id] = activity.id

        position = item.get("position")
        if position:
            layout_map[str(activity.id)] = position

    for item in transitions:
        item_id = item.get("id")
        input_id = item.get("input_id")
        output_id = item.get("output_id")
        if input_id in activity_map:
            input_id = activity_map[input_id]
        if output_id in activity_map:
            output_id = activity_map[output_id]
        if not input_id or not output_id:
            continue
        defaults = {
            "name": item.get("name") or "",
            "condition": item.get("condition") or "",
        }
        if item_id:
            transition = Transition.objects.get(id=item_id, process=process)
            transition.input_id = input_id
            transition.output_id = output_id
            transition.name = defaults["name"]
            transition.condition = defaults["condition"]
            transition.save()
        else:
            Transition.objects.create(
                process=process,
                input_id=input_id,
                output_id=output_id,
                name=defaults["name"],
                condition=defaults["condition"],
            )

    if begin_id in activity_map:
        begin_id = activity_map[begin_id]
    if end_id in activity_map:
        end_id = activity_map[end_id]

    if begin_id:
        process.begin_id = begin_id
    if end_id:
        process.end_id = end_id
    process.save()

    if positions:
        for key, value in positions.items():
            layout_map[str(key)] = value

    if layout:
        layout.layout = layout_map
        layout.save()
    else:
        WorkflowLayout.objects.create(process=process, layout=layout_map)

    return JsonResponse({"status": "ok"})




from django.template import Library
from django.conf import settings

register = Library()

@register.simple_tag
def graphics_url_prefix():
    """Returns the string contained in the setting GRAPHICS_JS_URL.
    
    (obsolete since django 1.0 ?)
    
    A suitable value is *http://www.walterzorn.com/scripts/*
    """
    return settings.GRAPHICS_JS_URL



from django.template import Library
from goflow.runtime.models import WorkItem
register = Library()

def mywork(user):
    '''
    Display the worklist of a user as a table.

    The template *goflow/workitems.html* is used for rendering.

    Usage::

        {% mywork user %}
    
    settings required::

        MIDDLEWARE_CLASSES = (
            'django.middleware.common.CommonMiddleware',
            ...
        )

    **Tip**

    this can be used instead, if *workitems* variable is available
    in the context::

        {% include "goflow/workitems.html" %}
    
    when using *render* shortcut, pass the request to
    include the context processors::

        def some_view(request):
        # ...
        return render(request, 'my_template.html', my_data_dictionary)
    '''
    workitems = WorkItem.objects.list_safe(user=user, noauto=True)
    return {'workitems':workitems}
mywork = register.inclusion_tag("goflow/workitems.html")(mywork)




import logging


from django.conf import settings
# from django.db.models import get_model
import sys

try:
    _file_log = settings.LOGGING_FILE
    _LOG_FILE_NOTSET = False
except AttributeError as e:
    _LOG_FILE_NOTSET = True
    _file_log = 'workflow.log'

if settings.DEBUG:
    level=logging.DEBUG
else:
    level=logging.INFO

log_format='%(asctime)s %(levelname)s %(module)s.%(funcName)s: %(message)s'
# python 2.4 ?
if sys.version_info[:2]==(2,4):
    log_format='%(asctime)s %(levelname)s %(module)s: %(message)s'
# log_format='%(asctime)s %(levelname)s %(name)s.%(funcName)s: %(message)s'

logging.basicConfig(
    filename=_file_log,
    level=level, 
    format=log_format,
    datefmt="%Y-%m-%d %H:%M:%S"
)
log = logging.getLogger('goflow.common')
if _LOG_FILE_NOTSET:
     log.warning('settings.LOGGING_FILE not set; default is workflow.log')


class Log(object):
    def __init__(self, module):
        self.log = logging.getLogger(module)
        # self._event = get_model('runtime', 'Event').objects

    def __getattr__(self, name):
        try:
            return getattr(self.log, name)
        except AttributeError as e:
            return getattr(self, name)

    def __call__(self, *args):
        if len(args) == 0: 
            return
        elif len(args) == 1:
            self.log.info(args[0] )
        else:
            self.log.info(u" ".join([str(i) for i in args]))

    def event(self, msg, workitem):
        # self._event.create(name=msg, workitem=workitem)
        self.log.info('EVENT: [%s] %s' % (str(workitem), msg))



#!/usr/local/bin/python
# -*- coding: utf-8 -*-
from django.db import models
from django.contrib.auth.models import Group, Permission
from django.contrib.contenttypes.models import ContentType
from django.urls.base import resolve

from django.conf import settings

from django import forms
# from admin_decorators import allow_tags
from django.utils.safestring import mark_safe

from datetime import timedelta
from django.utils import timezone
import logging

log = logging.getLogger('goflow.workflow.managers')


class Activity(models.Model):
    """Activities represent any kind of action an employee might want to do on an instance.

    The action might want to change the object instance, or simply
    route the instance on a given path. Activities are the places
    where any of these action are resolved by employees.a
    """
    KIND_CHOICES = (
                    ('standard', 'standard'),
                    ('dummy', 'dummy'),
                    ('subflow', 'subflow'),
                    )
    COMP_CHOICES = (
                    ('and', 'and'),
                    ('xor', 'xor'),
                    )
    title = models.CharField(max_length=100)
    kind = models.CharField(max_length=10, choices=KIND_CHOICES, verbose_name='type', default='standard')
    process = models.ForeignKey('Process', on_delete=models.CASCADE, related_name='activities')
    push_application = models.ForeignKey(
        'PushApplication', related_name='push_activities', on_delete=models.CASCADE, null=True, blank=True)
    pushapp_param = models.CharField(max_length=100, null=True, blank=True,
                                help_text="parameters dictionary; example: {'username':'john'}")
    application = models.ForeignKey('Application', related_name='activities', on_delete=models.CASCADE, null=True, blank=True,
                                help_text='leave it blank for prototyping the process without coding')
    app_param = models.CharField(max_length=100, verbose_name='parameters',
                                 help_text='parameters dictionary', null=True, blank=True)
    subflow = models.ForeignKey('Process', related_name='parent_activities',
                                on_delete=models.CASCADE, null=True, blank=True)
    roles = models.ManyToManyField(Group, related_name='activities', blank=True)
    description = models.TextField(null=True, blank=True)
    autostart = models.BooleanField(default=False)
    autofinish = models.BooleanField(default=True)
    join_mode = models.CharField(max_length=3, choices=COMP_CHOICES, verbose_name='join mode', default='xor')
    split_mode = models.CharField(max_length=3, choices=COMP_CHOICES, verbose_name='split mode', default='and')

    def nb_input_transitions(self):
        ''' returns the number of inputing transitions.
        '''
        return Transition.objects.filter(output=self, process=self.process).count()

    def __str__(self):
        return '%s (%s)' % (self.title, self.process.title)

    class Meta:
        unique_together = (("title", "process"),)
        verbose_name = 'Activity'
        verbose_name_plural = 'Activities'


class ProcessManager(models.Manager):
    '''Custom model manager for Process
    '''

    # TODO: also not too happy about this one.
    def is_enabled(self, title):
        '''
        Determines given a title if a process is enabled or otherwise

        :rtype: bool

        usage::

            if Process.objects.is_enabled('leave1'):
                # do something

        '''
        return self.get(title=title).enabled

    def check_can_start(self, process_name, user):
        '''
        Checks whether a process is enabled and whether the user has permission
        to instantiate it; raises exceptions if not the case, returns None otherwise.

        @type process_name: string
        @param process_name: a name of a process. e.g. 'leave'
        @type user: User
        @param user: an instance of django.contrib.auth.models.User,
                     typically retrieved through a request object.
        @rtype:
        @return: passes silently if checks are met,
                 raises exceptions otherwise.
        '''
        if not self.is_enabled(process_name):
            raise Exception('process %s disabled.' % process_name)

        if not user.is_superuser:
            if user.has_perm("workflow.can_instantiate"):
                lst = user.groups.filter(name=process_name)
                if lst.count() == 0 or \
                   (lst[0].permissions.filter(codename='can_instantiate').count() == 0):
                    raise Exception('permission needed to instantiate process %s.' % process_name)
            else:
                raise Exception('permission needed.')
        return


class Process(models.Model):
    """A process holds the map that describes the flow of work.

    The process map is made of activities and transitions.
    The instances you create on the map will begin the flow in
    the configured begin activity. Instances can be moved
    forward from activity to activity, going through transitions,
    until they reach the End activity.
    """
    enabled = models.BooleanField(default=True)
    date = models.DateTimeField(auto_now=True)
    title = models.CharField(max_length=100)
    description = models.TextField()
    # description.allow_tags=True
    begin = models.ForeignKey('Activity', related_name='bprocess', on_delete=models.CASCADE,
                              verbose_name='initial activity', null=True, blank=True)
    end = models.ForeignKey('Activity', related_name='eprocess', on_delete=models.CASCADE, verbose_name='final activity', null=True, blank=True,
                            help_text='a default end activity will be created if blank')
    priority = models.IntegerField(default=0)

    # add new ProcessManager
    objects = ProcessManager()

    class Meta:
        verbose_name_plural = 'Processes'
        permissions = (
            ("can_instantiate", "Can instantiate"),
            ("can_browse", "Can browse"),
        )

    def __str__(self):
        return self.title

    # @allow_tags
    def summary(self):
        return mark_safe('<pre>%s</pre>' % self.description)

    # @allow_tags
    def action(self):
        return mark_safe('add <a href=../activity/add/>a new activity</a> or <a href=../activity/>copy</a> an activity from another process')

    def add_activity(self, name):
        '''
        name: name of activity (get or created)
        '''
        a, created = Activity.objects.get_or_create(title=name, process=self,
                                                    defaults={'description':'(add description)'})
        return a

    def add_transition(self, name, activity_out, activity_in):
        t, created = Transition.objects.get_or_create(name=name, process=self,
                                             output=activity_out,
                                             defaults={'input':activity_in})
        return t

    def create_authorized_group_if_not_exists(self):
        if not getattr(settings, "GOFLOW_AUTO_CREATE_PROCESS_GROUPS", True):
            return
        g, created = Group.objects.get_or_create(name=self.title)
        if created:
            ptype = ContentType.objects.get_for_model(Process)
            cip = Permission.objects.get(content_type=ptype, codename='can_instantiate')
            g.permissions.add(cip)

    def save(self, force_insert=False, force_update=False, no_end=False, using=None,
             update_fields=None):
        models.Model.save(self)
        # instantiation group
        self.create_authorized_group_if_not_exists()

        if not no_end and not self.end:
            self.end = Activity.objects.create(title='End', process=self, kind='dummy', autostart=True)
            models.Model.save(self)
        try:
            if self.end and self.end.process.id != self.id:
                a = self.end
                a.id = None
                a.process = self
                a.save()
                self.end = a
                models.Model.save(self)
            if self.begin and self.begin.process.id != self.id:
                a = self.begin
                a.id = None
                a.process = self
                a.save()
                self.begin = a
                models.Model.save(self)
        except Exception:
            # admin console error ?!?
            pass


class WorkflowLayout(models.Model):
    process = models.OneToOneField(Process, on_delete=models.CASCADE, related_name="layout")
    layout = models.JSONField(default=dict, blank=True)
    updated_at = models.DateTimeField(auto_now=True)


class Application(models.Model):
    """ An application is a python view that can be called by URL.

        Activities can call applications.
        A commmon prefix may be defined: see settings.WF_APPS_PREFIX
    """
    url = models.CharField(max_length=255, unique=True, help_text='relative to prefix in settings.WF_APPS_PREFIX')
    # TODO: drop abbreviations (perhaps here not so necessary to ??
    SUFF_CHOICES = (
                    ('w', 'workitem.id'),
                    ('i', 'instance.id'),
                    ('o', 'object.id'),
                    )
    suffix = models.CharField(max_length=1, choices=SUFF_CHOICES, verbose_name='suffix', null=True, blank=True,
                               default='w', help_text='http://[host]/[settings.WF_APPS_PREFIX/][url]/[suffix]')
    detected_as_auto = None

    def __str__(self):
        return self.url

    def get_app_url(self, workitem=None, extern_for_user=None):
        from django.conf import settings
        path = '%s/%s/' % (settings.WF_APPS_PREFIX, self.url)
        if workitem:
            if self.suffix:
                if self.suffix == 'w': path = '%s%d/' % (path, workitem.id)
                if self.suffix == 'i': path = '%s%d/' % (path, workitem.instance.id)
                if self.suffix == 'o': path = '%s%d/' % (path, workitem.instance.content_object.id)
            else:
                path = '%s?workitem_id=%d' % (path, workitem.id)
        if extern_for_user:
            profile = UserProfile.objects.get(user=extern_for_user)
            path = 'http://%s%s' % (profile.web_host, path)
        return path

    def get_handler(self):
        '''returns handler mapped to url.
        '''
        try:
            func, args, kwargs = resolve(self.get_app_url() + '0/')
            self.detected_as_auto = False
        except Exception:
            func, args, kwargs = resolve(self.get_app_url())
            self.detected_as_auto = True
        return func

    # @allow_tags
    def documentation(self):
        doc = ''
        if self.detected_as_auto:
            doc = 'detected as auto application.<hr>'
        try:
            doc += u'<pre>%s</pre>' % self.get_handler().__doc__
        except Exception as v:
            doc += 'WARNING: the url %s is not resolved.' % self.get_app_url()
        return mark_safe(doc)

    def has_test_env(self):
        if Process.objects.filter(title='test_%s' % self.url).count() > 0:
            return True
        return False

    def create_test_env(self, user=None):
        if self.has_test_env():
            return

        g = Group.objects.create(name='test_%s' % self.url)
        ptype = ContentType.objects.get_for_model(Process)
        cip = Permission.objects.get(content_type=ptype, codename='can_instantiate')
        g.permissions.add(cip)
        # group added to current user
        if user:
            user.groups.add(g)

        p = Process.objects.create(title='test_%s' % self.url, description='unit test process')
        p.begin = p.end
        p.begin.autostart = False
        p.begin.title = "test_activity"
        p.begin.kind = 'standard'
        p.begin.application = self
        p.begin.description = 'test activity for application %s' % self.url
        p.begin.save()

        p.begin.roles.add(g)

        p.save()

    def remove_test_env(self):
        if not self.has_test_env(): return
        Process.objects.filter(title='test_%s' % self.url).delete()
        Group.objects.filter(name='test_%s' % self.url).delete()

    # @allow_tags
    def test(self):
        if self.has_test_env():
            return mark_safe(('<a href=teststart/%d/>start test instances</a> | '
                    '<a href=testenv/remove/%d/>remove unit test env</a>') % (self.id, self.id))
        else:
            return mark_safe('<a href=testenv/create/%d/>create unit test env</a>' % self.id)


class PushApplication(models.Model):
    """A push application routes a workitem to a specific user.
    It is a python function with the same prototype as the built-in
    one below::

        def route_to_requester(workitem):
            return workitem.instance.user

    Other parameters may be added (see Activity.pushapp_param field).
    Built-in push applications are implemented in pushapps module.
    A commmon prefix may be defined: see settings.WF_PUSH_APPS_PREFIX

    """
    url = models.CharField(max_length=255, unique=True)

    def get_handler(self):
        '''returns handler mapped to url.
        '''
        try:
            # search first in pre-built handlers
            from . import pushapps
            if self.url in dir(pushapps):
                # TODO
                return eval('pushapps.%s' % self.url)
            # then search elsewhere
            prefix = settings.WF_PUSH_APPS_PREFIX
            # dyn import
            #print("%s.%s" % (prefix, self.url))
            exec('import %s' % prefix)
           
            return eval('%s.%s' % (prefix, self.url))
        except Exception as v:
            log.error('PushApplication.get_handler %s', v)
        return None

    # @allow_tags
    def documentation(self):
        return mark_safe(u'<pre>%s</pre>' % self.get_handler().__doc__)

    def execute(self, workitem, **kwargs):
        handler = self.get_handler()
        return handler(workitem, **kwargs)

    def __str__(self):
        return self.url

    # @allow_tags
    def test(self):
        return mark_safe('<a href=#>test (not yet implemented)</a>')


class Transition(models.Model):
    """ A Transition connects two Activities: a From and a To activity.

    Since the transition is oriented you can think at it as being a
    link starting from the From and ending in the To activity.
    Linking the activities in your process you will be able to draw
    the map.

    Each transition is associated to a condition that will be tested
    each time an instance has to choose which path to follow.
    If the only transition whose condition is evaluated to true will
    be the transition choosen for the forwarding of the instance.
    """
    name = models.CharField(max_length=50, null=True, blank=True)
    process = models.ForeignKey(
        Process, on_delete=models.CASCADE, related_name='transitions')
    input = models.ForeignKey(
        Activity, on_delete=models.CASCADE, related_name='transition_inputs')
    condition = models.CharField(max_length=200, null=True, blank=True,
                                 help_text='ex: instance.condition=="OK" | OK')
    output = models.ForeignKey(
        Activity, on_delete=models.CASCADE, related_name='transition_outputs')
    description = models.CharField(max_length=100, null=True, blank=True)
    precondition = models.SlugField(null=True, blank=True, help_text='object method that return True if transition is posible')

    def is_transition(self):
        ''' used in admin templates.
        '''
        return True

    def save(self):
        if self.input.process != self.process or self.output.process != self.process:
            raise Exception("a transition and its activities must be linked to the same process")
        models.Model.save(self)

    def __str__(self):
        return self.name or 't%s' % str(self.pk)

    class Meta:
        pass
        # unique_together = (("input", "condition"),)


class UserProfile(models.Model):
    """Contains workflow-specific user data.
    """
    user = models.OneToOneField(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    web_host = models.CharField(max_length=100, default='localhost:8000')
    notified = models.BooleanField(default=True, verbose_name='notification by email')
    last_notif = models.DateTimeField()
    nb_wi_notif = models.IntegerField(default=1, verbose_name='items before notification',
                                      help_text='notification if the number of items waiting is reached')

    notif_delay = models.IntegerField(default=1, verbose_name='Notification delay', help_text='in days')
    urgent_priority = models.IntegerField(default=5, verbose_name='Urgent priority threshold', help_text='a mail notification is sent when an item has at least this priority level')

    def save(self, **kwargs):
        if not self.last_notif: self.last_notif = timezone.now()
        models.Model.save(self, **kwargs)

    def check_notif_to_send(self):
        now = timezone.now()
        if now > self.last_notif + timedelta(days=self.notif_delay or 1):
            return True
        return False

    def notif_sent(self):
        now = timezone.now()
        self.last_notif = now
        self.save()

    class Meta:
        verbose_name = 'Workflow user profile'
        verbose_name_plural = 'Workflow users profiles'




from django.template import Library
#from django.conf import settings
from goflow.apptools.models import ImageButton

register = Library()

def form_ext(form):
    '''
    This will insert a form, in a bit more sophisticated way than {{ form }}.

    Required and optional fields are displayed differently. For details,
    see template file *goflow/apptools/edit_form.html*.

    parameter:

    form
        form variable

    Usage::

        {% form_ext form %}
    
    the current implementation is equivalent to::

        {% include "goflow/apptools/edit_form.html" %}
    '''
    return {'form':form}
form_ext = register.inclusion_tag("goflow/apptools/edit_form.html")(form_ext)

def _get_transitions_out_images(activity):
    if activity.split_mode == 'and':
        raise Exception('_get_transitions_out_images: xor split_mode required')
    transitions = Transition.objects.filter(input=activity)
    icons = []
    for t in transitions:
        if t.icon: icons.append(t.icon)


def input_buttons(context):
    '''
    This will insert submit buttons in application templates.

    The template *goflow/apptools/input_buttons.html* is used for rendering.

    Usage::

        {% input_buttons %}

    - *Context requirements*

    submit_name
        name of the buttons
    ok_values
        sequence of values, used for button value generation
    cancel_value
        value for the *cancel* button

    - *Context options*

    save_value
        can be used with the *edit_model* application: this button
        saves the object but does not complete the task.
    '''
    sub_context ={'submit_name':context['submit_name']}
    sub_context['ok_values'] = context.get('ok_values')
    sub_context['save_value'] = context.get('save_value')
    sub_context['cancel_value'] = context.get('cancel_value')
    return sub_context
input_buttons = register.inclusion_tag("goflow/apptools/input_buttons.html", takes_context=True)(input_buttons)

@register.simple_tag
def image_button(action):
    '''
    This will insert submit image buttons in application templates.
    
    Usage::

        {% for value in ok_values %}
            {% image_button value %}
        {% endfor %}
        
        {% image_button cancel_value %}

    ok_values
        sequence of actions corresponding to transition conditions.
        Example: when the image button with the *edit* action is pressed,
        the transition such as *transition.condition = 'edit'* will be
        cadidate for workitem forwarding.

    cancel_value
        action correponding to cancel.
    
    An *action* (like *edit*, *new*, *exit*, ...) is a "slug" mapped to
    an image, or more exactly an *Icon* model; this mapping is implemented
    with *ImageButton* model.
    '''
    return ImageButton.objects.get(action=action).graphic_input()



#!/usr/local/bin/python
# -*- coding: utf-8 -*-

from django.template import Context, Template
from django.template.loader import render_to_string
def send_mail(workitems=None, user=None, subject='message', template='mail.txt'):
    # subject
    try:
        subject = settings.EMAIL_SUBJECT_PREFIX + subject
    except Exception:
        pass
    t = Template(subject)
    subject = t.render(Context({'workitems': workitems,'user':user}))
    from goflow.workflow.models import UserProfile
    profile = UserProfile.objects.get(user=user) 
    message = render_to_string(template, {
                                          'workitems': workitems,
                                          'user':user,
                                          'url_prefix':'http://%s/' % profile.web_host
                                          })
    
    user.email_user(subject, message)



#!/usr/local/bin/python
# -*- coding: utf-8 -*-
from django.db import models
from django.contrib.contenttypes.models import ContentType
from goflow.workflow.models import Process, Activity
#from goflow.workflow.decorators import allow_tags
from django.utils.safestring import mark_safe

class ProcessImage(models.Model):
    process = models.ForeignKey(Process, on_delete=models.CASCADE)
    file = models.ImageField(upload_to='images')
    
    #@allow_tags
    def graphic(self):
        return mark_safe('<img name=image%d src=%s>' % (self.id, self.file.url))

    #@allow_tags
    def graphic_input(self):
        return mark_safe('<input type=image name=process src=%s>' % self.file.url)
    
    def list_activities(self):
        return self.process.activities.all()
    
    def list_activity_positions(self):
        return ActivityPosition.objects.filter(diagram=self)
        
    def __str__(self):
        return self.process.title

class ActivityPosition(models.Model):
    diagram = models.ForeignKey(ProcessImage, on_delete=models.CASCADE)
    activity = models.ForeignKey(Activity, on_delete=models.CASCADE)
    x = models.PositiveSmallIntegerField(default=0)
    y = models.PositiveSmallIntegerField(default=0)



#!/usr/local/bin/python
# -*- coding: utf-8 -*-
from django.db import models
from django.contrib.auth.models import Group
from goflow.workflow.models import Process, Activity, Transition, UserProfile
from goflow.workflow.notification import send_mail
from datetime import timedelta, datetime
from django.urls.base import resolve
from django.core.mail import mail_admins

from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes import fields
from django.contrib.auth import get_user_model

from goflow.workflow.logger import Log; log = Log('goflow.runtime.managers')
from django.conf import settings

# from goflow.workflow.decorators import allow_tags
from django.utils.safestring import mark_safe

User = get_user_model()


class ProcessInstanceManager(models.Manager):
    '''Custom model manager for ProcessInstance
    '''
   
    def start(self, process_name, user, item, title=None, priority=0):
        '''
        Returns a workitem given the name of a preexisting enabled Process 
        instance, while passing in the id of the user, the contenttype 
        object and the title.
        
        :type process_name: string
        :param process_name: a name of a process. e.g. 'leave'
        :type user: User
        :param user: an instance of django.contrib.auth.models.User, 
                     typically retrieved through a request object.
        :type item: ContentType
        :param item: a content_type object e.g. an instance of LeaveRequest
        :type: title: string
        :param title: title of new ProcessInstance instance (optional)
        :type: priority: integer
        :param priority: default priority (optional)
        :rtype: WorkItem
        :return: a newly configured workitem sent to auto_user, 
                 a target_user, or ?? (roles).
        
        usage::
            
            wi = Process.objects.start(process_name='leave', 
                                       user=admin, item=leaverequest1)

        '''
        process = Process.objects.get(title=process_name, enabled=True)
        if priority == 0: priority = process.priority
            
        if not title or (title == 'instance'):
            title = '%s %s' % (process_name, str(item))
        instance = ProcessInstance.objects.create(process=process, user=user, title=title, content_object=item)
        # instance running
        instance.set_status('running')
        
        workitem = WorkItem.objects.create(instance=instance, user=user,
                                           activity=process.begin, priority=priority)
        log.event('created by ' + user.username, workitem)
        log('process:', process_name, 'user:', user.username, 'item:', item.id)
    
        if process.begin.kind == 'dummy':
            log('routing activity', process.begin.title, 'workitem:', workitem)
            auto_user = User.objects.get(username=settings.WF_USER_AUTO)
            workitem.activate(actor=auto_user)
            workitem.complete(actor=auto_user)
            return workitem
        
        if process.begin.autostart:
            log('run auto activity', process.begin.title, 'workitem:', workitem)
            auto_user = User.objects.get(username=settings.WF_USER_AUTO)
            workitem.activate(actor=auto_user)
    
            if workitem.exec_auto_application():
                log('workitem.exec_auto_application:', workitem)
                workitem.complete(actor=auto_user)
            return workitem

        if process.begin.push_application:
            target_user = workitem.exec_push_application()
            log('application pushed to user', target_user.username)
            workitem.user = target_user
            workitem.save()
            log.event('assigned to ' + target_user.username, workitem)
            # notify_if_needed(user=target_user)
        else:
            # set pull roles; useful (in activity too)?
            workitem.pull_roles.set(workitem.activity.roles.all())
            workitem.save()
            # notify_if_needed(roles=workitem.pull_roles)
        
        return workitem


class ProcessInstance(models.Model):
    """ This is a process instance.
    
    A process instance is created when someone decides to do something,
    and doing this thing means start using a process defined in GoFlow.
    That's why it is called "process instance". The process is a class
    (=the definition of the process), and each time you want to
    "do what is defined in this process", that means you want to create
    an INSTANCE of this process.
    
    So from this point of view, an instance represents your dynamic
    part of a process. While the process definition contains the map
    of the workflow, the instance stores your usage, your history,
    your state of this process.
    
    The ProcessInstance will collect and handle workitems (see definition)
    to be passed from activity to activity in the process.
    
    Each instance can have more than one workitem depending on the
    number of split actions encountered in the process flow.
    That means that an instance is actually the collection of all of
    the instance "pieces" (workitems) that we get from splits of the
    same original process instance.
    
    Each ProcessInstance keeps track of its history through a graph.
    Each node of the graph represents an activity the instance has
    gone through (normal graph nodes) or an activity the instance is
    now pending on (a graph leaf node). Tracking the ProcessInstance history
    can be very useful for the ProcessInstance monitoring.
    
    When a process instance starts, the instance has to carry an
    implementation object that contains the application data. The
    specifications for the implementation class is:
    
    (nothing: now managed by generic relation)
    
    From the instance, the implementation object is reached as following:
      obj = instance.content_object (or instance.wfobject()).
    In a template, a field date1 will be displayed like this:
      {{ instance.wfobject.date1 }} or {{ instance.content_object.date1 }}
    
    From the object, instances may be reached with the reverse generic relation:
    the following can be added to the model:
      wfinstances = generic.GenericRelation(ProcessInstance)
    
    """
    STATUS_CHOICES = (
                      ('initiated', 'initiated'),
                      ('running', 'running'),
                      ('active', 'active'),
                      ('complete', 'complete'),
                      ('terminated', 'terminated'),
                      ('suspended', 'suspended'),
                      )
    title = models.CharField(max_length=100)
    process = models.ForeignKey(
        Process, related_name='instances', on_delete=models.CASCADE, null=True, blank=True)
    creationTime = models.DateTimeField(auto_now_add=True)
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.CASCADE, related_name='instances')
    status = models.CharField(max_length=10, choices=STATUS_CHOICES, default='initiated')
    old_status = models.CharField(max_length=10, choices=STATUS_CHOICES, null=True, blank=True)
    condition = models.CharField(max_length=50, null=True, blank=True)
    
    # refactoring
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.PositiveIntegerField()
    content_object = fields.GenericForeignKey('content_type', 'object_id')
    
    # add new ProcessInstanceManager
    objects = ProcessInstanceManager()
    
    def wfobject(self):
        return self.content_object
    
    # @allow_tags
    def workitems_list(self):
        '''provide html link to workitems for a process instance in admin change list.
        @rtype: string
        @return: html href link "../workitem/?instance__id__exact=[self.id]&ot=asc&o=0"
        '''
        nbwi = self.workitems.count()
        return mark_safe('<a href=../workitem/?instance__id__exact=%d&ot=asc&o=0>%d item(s)</a>' % (self.pk, nbwi))
    
    def __str__(self):
        return self.title
    
    def set_status(self, status):
        if not status in [x for x, y in ProcessInstance.STATUS_CHOICES]:
            raise Exception('instance status incorrect :%s' % status)
        self.old_status = self.status
        self.status = status
        self.save()


class WorkItemManager(models.Manager):
    '''Custom model manager for WorkItem
    '''

    def get_safe(self, id, user=None, enabled_only=False, status=('inactive', 'active')):
        '''
        Retrieves a single WorkItem instance given a set of parameters
        
        :type id: int
        :param id: the id of the WorkItem instance
        :type user: User
        :param user: an instance of django.contrib.auth.models.User, 
                     typically retrieved through a request object.
        :type enabled_only: bool
        :param enabled_only: implies that only enabled processes should be queried
        :type status: tuple or string
        :param status: ensure that workitem has one of the given set of statuses
        
        usage::
        
            workitem = WorkItem.objects.get_safe(id, user=request.user)
        
        '''
        if enabled_only:
            workitem = self.get(id=id, activity__process__enabled=True)
        else:
            workitem = self.get(id=id)
        workitem._check(user, status)
        return workitem

    def list_safe(self, user=None, username=None, queryset='qs_default', activity=None, status=None,
                      notstatus=('blocked', 'suspended', 'fallout', 'complete'), noauto=True):
        """
        Retrieve list of workitems (in order to display a task list for example).
        
        :type user: User
        :param user: filter on instance of django.contrib.auth.models.User (default=all) 
        :type username: string
        :param username: filter on username of django.contrib.auth.models.User (default=all) 
        :type queryset: QuerySet
        :param queryset: pre-filtering (default=WorkItem.objects)
        :type activity: Activity
        :param activity: filter on instance of goflow.workflow.models.Activity (default=all) 
        :type status: string
        :param status: filter on status (default=all) 
        :type notstatus: string or tuple
        :param notstatus: list of status to exclude (default: [blocked, suspended, fallout, complete])
        :type noauto: bool
        :param noauto: if True (default) auto activities are excluded.
        
        usage::
        
            workitems = WorkItem.objects.list_safe(user=me, notstatus='complete', noauto=True)
        
        """
        if queryset == 'qs_default': queryset = WorkItem.objects
        if status: notstatus = []
        
        groups = Group.objects.all()
        if user:
            query = queryset.filter(user=user, activity__process__enabled=True).order_by('-priority')
            groups = user.groups.all()
        else:
            if username:
                query = queryset.filter(
                        user__username=username,
                        activity__process__enabled=True
                ).order_by('-priority')
                groups = User.objects.get(username=username).groups.all()
            else:
                query = None
        if query:
            if status:
                query = query.filter(status=status)
            
            if notstatus:
                for s in notstatus: 
                    query = query.exclude(status=s)
            
            if noauto:
                query = query.exclude(activity__autostart=True)
            
            if activity:
                # TODO: this is not used...??
                sq = query.filter(activity=activity)
            
            query = list(query)
        else:
            query = []
        
        # search pullable workitems
        for role in groups:
            pullables = queryset.filter(pull_roles=role, activity__process__enabled=True).order_by('-priority')
            if status:
                pullables = pullables.filter(status=status)
            
            if notstatus:
                for s in notstatus:
                    pullables = pullables.exclude(status=s)
            
            if noauto:
                pullables = pullables.exclude(activity__autostart=True)
            
            if activity:
                pullables = pullables.filter(activity=activity)
            
            if user:
                pullables = pullables.filter(user__isnull=True)  # tricky
                pullables = pullables.exclude(user=user)
                query.extend(list(pullables))
            
            if username:
                pullables = pullables.exclude(user__username=username)
            
            log.debug('pullables workitems role %s: %s', role, str(pullables))
            query.extend(list(pullables))
        
        # search workitems pullable by anybody
        pullables = queryset.filter(pull_roles__isnull=True,
                                    activity__process__enabled=True,
                                    user__isnull=True).order_by('-priority')
        if status:
            pullables = pullables.filter(status=status)
        if notstatus:
            for s in notstatus:
                pullables = pullables.exclude(status=s)
        if noauto:
            pullables = pullables.exclude(activity__autostart=True)
        if activity:
            pullables = pullables.filter(activity=activity)
        if pullables.count() > 0:
            log.debug('anybody\'s workitems: %s', str(pullables))
            query.extend(list(pullables))
        
        return query
    
    def notify_if_needed(self, user=None, roles=None):
        ''' notify user if conditions are fullfilled
        '''
        if user:
            workitems = self.list_safe(user=user, notstatus='complete', noauto=True)
            profile, created = UserProfile.objects.get_or_create(user=user)
            if len(workitems) >= profile.nb_wi_notif:
                try:
                    if profile.check_notif_to_send():
                        send_mail(workitems=workitems, user=user, subject='message', template='mail.txt')
                        profile.notif_sent()
                        log.info('notification sent to %s' % user.username)
                except Exception as v:
                    log.error('sendmail error: %s' % v)
        return


class WorkItem(models.Model):
    """A workitem object represents an activity you are performing.
    
    An Activity object defines the activity, while the workitem object
    represents that you are performing this activity. So workitem is
    an "instance" of the activity.
    """
    STATUS_CHOICES = (
                      ('blocked', 'blocked'),
                      ('inactive', 'inactive'),
                      ('active', 'active'),
                      ('suspended', 'suspended'),
                      ('fallout', 'fallout'),
                      ('complete', 'complete'),
                      )
    date = models.DateTimeField(auto_now=True)
    user = models.ForeignKey(settings.AUTH_USER_MODEL, related_name='workitems',
                             on_delete=models.CASCADE, null=True, blank=True)
    instance = models.ForeignKey(
        ProcessInstance, on_delete=models.CASCADE, related_name='workitems')
    activity = models.ForeignKey(
        Activity, on_delete=models.CASCADE, related_name='workitems')
    workitem_from = models.ForeignKey(
        'self', related_name='workitems_to', on_delete=models.CASCADE, null=True, blank=True)
    others_workitems_from = models.ManyToManyField('self', blank=True)
    push_roles = models.ManyToManyField(Group, related_name='push_workitems', blank=True)
    pull_roles = models.ManyToManyField(Group, related_name='pull_workitems', blank=True)
    blocked = models.BooleanField(default=False)
    priority = models.IntegerField(default=0)
    status = models.CharField(max_length=10, choices=STATUS_CHOICES, default='inactive')

    objects = WorkItemManager()
    
    def forward(self, timeout_forwarding=False, subflow_workitem=None):
        # forward_workitem(workitem, path=None, timeout_forwarding=False, subflow_workitem=None):
        '''
        Convenience procedure to forwards workitems to valid destination activities.
        
        @type path: string??
        @param path: XXX TODO: This is not used, so don't know why it's here.
        @type timeoutForwarding: bool
        @param timeoutForwarding:
        @type: subflow_workitem: WorkItem
        @param subflow_workitem: a workitem associated with a subflow ???
        
        '''
        log.info('forward_workitem %s', str(self))
        if not timeout_forwarding:
            if self.status != 'complete':
                return
        if self.has_workitems_to() and not subflow_workitem:
            log.debug('forward_workitem canceled for %s: ' 
                       'workitem.has_workitems_to()', str(self))
            return
        
        if timeout_forwarding:
            log.info('timeout forwarding')
            Event.objects.create(name='timeout', workitem=self)
        
        for destination in self.get_destinations(timeout_forwarding):
            self._forward_workitem_to_activity(destination)
            if self.activity.split_mode == 'xor': break

    def _forward_workitem_to_activity(self, target_activity):
        '''
        Passes the process instance embedded in the given workitem 
        to a new workitem that is associated with the destination activity.
        
        @type target_activity: Activity
        @param target_activity: the activity instance to which the workitem 
                                should be forwarded
        @rtype: WorkItem
        @return: a workitem that has been passed on to the next 
                 activity (and next user)
        '''
        instance = self.instance
        # search a blocked workitem first
        qwi = WorkItem.objects.filter(instance=instance, activity=target_activity, status='blocked')
        if qwi.count() == 0:
            wi = WorkItem.objects.create(instance=instance, activity=target_activity,
                                         user=None, priority=self.priority)
            created = True
            log.info('forwarded to %s', target_activity.title)
            Event.objects.create(name='creation by %s' % self.user.username, workitem=wi)
            Event.objects.create(name='forwarded to %s' % target_activity.title, workitem=self)
            wi.workitem_from = self
        else:
            created = False
            wi = qwi[0]
        
        if target_activity.join_mode == 'and':
            nb_input_transitions = target_activity.nb_input_transitions()
            if nb_input_transitions > 1:
                if created:
                    # first worktem: block it
                    wi.block()
                    return    
                else:
                    wi.others_workitems_from.add(self)
                    if wi.others_workitems_from.all().count() + 1 < nb_input_transitions:
                        # keep blocked
                        return
                    else:
                        # check if the join is OK
                        if wi.check_join():
                            wi.status = 'inactive'
                            wi.save()
                            log.info('activity %s: workitem %s unblocked', target_activity.title, str(wi))
                        else:
                            return
        else:
            if not created:
                # join_mode='and'
                log.error('activity %s: join_mode must be and', target_activity.title)
                self.fall_out()
                wi.fall_out()
                return
        
        if target_activity.autostart:
            log.info('run auto activity %s workitem %s', target_activity.title, str(wi))
            try:
                auto_user = User.objects.get(username=settings.WF_USER_AUTO)
            except Exception:
                error = 'a user named %s (settings.WF_USER_AUTO) must be defined for auto activities'
                raise Exception(error % settings.WF_USER_AUTO)
            wi.activate(actor=auto_user)
            if wi.exec_auto_application():
                wi.complete(actor=auto_user)
            return wi
        
        if target_activity.push_application:
            target_user = wi.exec_push_application()
            log.info('application pushed to user %s', target_user.username)
            wi.user = target_user
            wi.save()
            Event.objects.create(name='assigned to %s' % target_user.username, workitem=wi)
            WorkItem.objects.notify_if_needed(user=target_user)
        else:
            wi.pull_roles = wi.activity.roles.all()
            wi.save()
            WorkItem.objects.notify_if_needed(roles=wi.pull_roles)
        return wi
    
    def check_join(self):
        log.warning('workitem check_join NYI- useful ?')
        return True
    
    def _check(self, user, status=('inactive', 'active')):
        '''
        helper function to determine whether process is:
            - enabled, etc..
        
        '''
        if type(status) == type(''):
            status = (status,)
            
        if not self.activity.process.enabled:
            error = 'process %s disabled.' % self.activity.process.title
            log.error('workitem._check: %s' % error)
            raise Exception(error)
            
        if not self.check_user(user):
            error = 'user %s cannot take workitem %d.' % (user.username, self.pk)
            log.error('workitem._check: %s' % error)
            self.fall_out()
            raise Exception(error)
            
        if not self.status in status:
            error = 'workitem %d has not a correct status (%s/%s).' % (
                self.pk, self.status, str(status))
            log.error('workitem._check: %s' % error)
            raise Exception(error)
        return
    
    def get_destinations(self, timeout_forwarding=False):
        # get_destinations(workitem, path=None, timeout_forwarding=False):
        '''
        Return list of destination activities that meet the conditions of each transition
        
        @type path: string??
        @param path: XXX TODO: This is not used, so don't know why it's here.
        @type timeout_forwarding: bool
        @param timeout_forwarding: a workitem with a time-delay??
        @rtype: [Activity]
        @return: list of destination activities.
        '''
        transitions = Transition.objects.filter(input=self.activity)
        if timeout_forwarding:
            transitions = transitions.filter(condition__contains='workitem.time_out')
        destinations = []
        for t in transitions:
            if self.eval_transition_condition(t):
                destinations.append(t.output)
        return destinations
    
    def eval_transition_condition(self, transition):
        '''
        evaluate the condition of a transition
        '''
        if not transition.condition:
            return True
        instance = self.instance
        wfobject = instance.wfobject()
        log.debug('eval_transition_condition %s - %s',
            transition.condition, instance.condition)
        try:
            result = eval(transition.condition)
            
            # boolean expr
            if type(result) == type(True):
                log.debug('eval_transition_condition boolean %s', str(result))
                return result
            if type(result) == type(''):
                log.debug('eval_transition_condition cmp instance.condition %s', str(instance.condition == result))
                return (instance.condition == result)
        except Exception as v:
            log.debug('eval_transition_condition [%s]: %s', transition.condition, v)
            return (instance.condition == transition.condition)
            # log.error('eval_transition_condition [%s]: %s', transition.condition, v)
        return False
    
    def exec_push_application(self):
        '''
        Execute push application in workitem
        '''
        if not self.activity.process.enabled:
            raise Exception('process %s disabled.' % self.activity.process.title)
        params = self.activity.pushapp_param
        try:
            if params: kwargs = eval(params)
            else: kwargs = {}

            #print("Test")
            result = self.activity.push_application.execute(self, **kwargs)
        except Exception as v:
            log.error('exec_push_application %s', v)
            result = None
            self.fall_out()
        return result
    
    def exec_auto_application(self):
        '''
        creates a test auto application for activities that don't yet have applications
        @rtype: bool
        '''
        try:
            if not self.activity.process.enabled:
                raise Exception('process %s disabled.' % self.activity.process.title)
            # no application: default auto app
            if not self.activity.application:
                return self.default_auto_app()
            
            func, args, kwargs = resolve(self.activity.application.get_app_url())
            params = self.activity.app_param
            # params values defined in activity override those defined in urls.py
            if params:
                params = eval('{' + params.lstrip('{').rstrip('}') + '}')
                kwargs.update(params)
            func(workitem=self , **kwargs)
            return True
        except Exception as v:
            log.error('execution wi %s:%s', self, v)
        return False
    
    def default_auto_app(self):
        '''
        retrieves wfobject, logs info to it saves
        
        @rtype: bool
        @return: always returns True
        '''
        obj = self.instance.wfobject()
        obj.history += '\n>>> execute auto activity: [%s]' % self.activity.title
        obj.save()
        return True
    
    def activate(self, actor):
        '''
        changes workitem status to 'active' and logs event, activator
        
        '''
        self._check(actor, ('inactive', 'active'))
        if self.status == 'active':
            log.warning('activate_workitem actor %s workitem %s already active',
                        actor.username, str(self))
            return
        self.status = 'active'
        self.user = actor
        self.save()
        log.info('activate_workitem actor %s workitem %s',
                 actor.username, str(self))
        Event.objects.create(name='activated by %s' % actor.username, workitem=self)
    
    def complete(self, actor):
        '''
        changes status of workitem to 'complete' and logs event
        '''
        self._check(actor, 'active')
        self.status = 'complete'
        self.user = actor
        self.save()
        log.info('complete_workitem actor %s workitem %s', actor.username, str(self))
        Event.objects.create(name='completed by %s' % actor.username, workitem=self)
        
        if self.activity.autofinish:
            log.debug('activity autofinish: forward')
            self.forward()
        
        # if end activity, instance is complete
        if self.instance.process.end == self.activity:
            log.info('activity end process %s' % self.instance.process.title)
            # first test subflow
            lwi = WorkItem.objects.filter(activity__subflow=self.instance.process,
                                          status='blocked',
                                          instance=self.instance)
            if lwi.count() > 0:
                log.info('parent process for subflow %s' % self.instance.process.title)
                workitem0 = lwi[0]
                workitem0.instance.process = workitem0.activity.process
                workitem0.instance.save()
                log.info('process change for instance %s' % workitem0.instance.title)
                workitem0.status = 'complete'
                workitem0.save()
                workitem0.forward(subflow_workitem=self)
            else:
                self.instance.set_status('complete')
    
    def start_subflow(self, actor=None):
        '''
        starts subflow and blocks passed in workitem
        '''
        if not actor: actor = self.user
        subflow_begin_activity = self.activity.subflow.begin
        instance = self.instance
        instance.process = self.activity.subflow
        instance.save()
        self.status = 'blocked'
        self.blocked = True
        self.save()
        
        sub_workitem = self._forward_workitem_to_activity(subflow_begin_activity)
        return sub_workitem
    
    def eval_condition(self, transition):
        '''
        evaluate the condition of a transition
        '''
        raise Exception("New API (not yet implemented)")
    
    def __str__(self):
        return '%s-%s-%s' % (str(self.instance), str(self.activity), str(self.pk))
    
    def has_workitems_to(self):
        b = (self.workitems_to.count() > 0)
        return b
    
    def check_user(self, user):
        """returns True if authorized, False if not.
        
        For dummy activities, returns always True
        """
        if self.activity.kind == 'dummy':
            return True
        
        if user and self.user and self.user != user:
            return False
        ugroups = user.groups.all()
        agroups = self.activity.roles.all()
        authorized = False
        if agroups and len(agroups) > 0:
            for g in ugroups:
                if g in agroups:
                    authorized = True
                    break
        else:
            authorized = True
        return authorized
            
    def set_user(self, user, commit=True):
        """affect user if he has a role authorized for activity.
        
        return True if authorized, False if not (workitem then falls out)
        """
        if self.check_user(user):
            self.user = user
            if commit: self.save()
            return True
        self.fallOut()
        return False
    
    def can_priority_change(self):
        '''can the user change priority.
        
        @rtype: bool
        @return: returns True if the user can change priority
        
        the user must belong to a group with "workitem.can_change_priority"  permission,
        and this group's name must be the same as the process title.
        '''
        if self.user.has_perm("workitem.can_change_priority"):
            lst = self.user.groups.filter(name=self.instance.process.title)
            if lst.count() == 0 or \
               (lst[0].permissions.filter(codename='can_change_priority').count() == 0):
                return False
            return True
        return False
    
    def block(self):
        self.status = 'blocked'
        self.save()
        Event.objects.create(name='blocked', workitem=self)
    
    def fall_out(self):
        self.status = 'fallout'
        self.save()
        Event.objects.create(name='fallout', workitem=self)
        if not settings.DEBUG:
            mail_admins(subject='workflow workitem %s fall out' % str(self.pk),
                    message=u'''
The workitem [%s] was falling out.

Process:  %s
Activity: %s
instance: %s
----------------------------------
                    ''' % (
                           self.instance.process,
                           self.activity,
                           self.instance,
                           ))
    
    def html_action(self):
        label = 'action'
        if self.status == 'inactive':
            label = 'activate'
            url = 'activate/%d/' % self.id
        if self.status == 'active':
            label = 'complete'
            url = 'complete/%d/' % self.id
        if self.status == 'complete':
            return 'completed'
        return '<a href=%s>%s</a>' % (url, label)
    
    def html_action_link(self):
        if self.status == 'inactive':
            url = 'activate/%d/' % self.id
        if self.status == 'active':
            url = 'complete/%d/' % self.id
        if self.status == 'complete':
            raise Exception('no action for completed workitems')
        return url
    
    def time_out(self, delay, unit='days'):
        '''
        return True if timeout reached
          delay:    nb units
          unit: 'weeks' | 'days' | 'hours' ... (see timedelta)
        '''
        tdelta = eval('timedelta(' + unit + '=delay)')
        now = datetime.now()
        return (now > (self.date + tdelta))
    
    # @allow_tags
    def events_list(self):
        '''provide html link to events for a workitem in admin change list.
        @rtype: string
        @return: html href link "../event/?workitem__id__exact=[self.id]&ot=asc&o=0"
        '''
        nbevt = self.events.count()
        return mark_safe('<a href=../event/?workitem__id__exact=%d&ot=asc&o=0>%d item(s)</a>' % (self.pk, nbevt))
    
    class Meta:
        permissions = (
            ("can_change_priority", "Can change priority"),
        )


class Event(models.Model):
    """Event are changes that happens on workitems.
    """
    date = models.DateTimeField(auto_now=True)
    name = models.CharField(max_length=50)
    workitem = models.ForeignKey(
        WorkItem, on_delete=models.CASCADE, related_name='events')
    
    def __str__(self):
        return self.name




#!/usr/local/bin/python
# -*- coding: utf-8 -*-
from django.shortcuts import render
from django.http import HttpResponseRedirect

from goflow.graphics.models import Graph
from goflow.workflow.models import Process


def graph(request, id, template='goflow/graphics/graph.html'):
    processes = Process.objects.all()
    graph = Graph.objects.get(id=(int(id)))
    return render(request, template, {'processes':processes, 'graph':graph})


def graph_save(request, id):
    # save positions TODO
    return HttpResponseRedirect('..')



from django.urls import path

from goflow.graphics import views

urlpatterns = [
    path('<int:id>/save/', views.graph_save),
    path('<int:id>/', views.graph, {'template':'goflow/graphics/graph.html'}),
]



#!/usr/local/bin/python
# -*- coding: utf-8 -*-
from django.http import HttpResponseRedirect
from goflow.graphics2.models import ProcessImage, ActivityPosition
from goflow.workflow.models import Activity
from django.contrib import messages

def pos_activity(request, process_id):
    process = ProcessImage.objects.get(id=int(process_id))
    activity = Activity.objects.get(id=int(request.GET['activity']))
    activity_pos, created = ActivityPosition.objects.get_or_create(activity=activity, diagram=process)
    x = int(request.GET['process.x'])
    y = int(request.GET['process.y'])
    activity_pos.x = x
    activity_pos.y = y
    activity_pos.save()
    messages.add_message(request, messages.INFO,
                         'activity %s in positioned in the diagram of process %s.' % (
                             activity.title, process.process.title))
    #request.user.message_set.create(
    #    message='activity %s in positioned in the diagram of process %s.'% (activity.title, process.process.title)
    #)
    return HttpResponseRedirect('..')



#!/usr/local/bin/python
# -*- coding: utf-8 -*-
from django.shortcuts import render
from django.http import HttpResponse, HttpResponseRedirect

from django.contrib.auth.decorators import login_required

from goflow.runtime.models import WorkItem, ProcessInstance
from django.utils.translation import gettext_lazy as _, gettext


@login_required
def mywork(request, template='goflow/mywork.html'):
    '''
    displays the worklist of the current user.
    
    parameters:
    
    template
        default:'goflow/mywork.html'
    '''
    workitems = WorkItem.objects.list_safe(user=request.user, noauto=True)
    return render(request, template, {'workitems':workitems})


@login_required
def otherswork(request, template='goflow/otherswork.html'):
    worker = request.GET['worker']
    workitems = WorkItem.objects.list_safe(username=worker, noauto=False)
    return render(request, template, {'worker':worker, 'workitems':workitems})


@login_required
def instancehistory(request, template='goflow/instancehistory.html'):
    id = int(request.GET['id'])
    inst = ProcessInstance.objects.get(pk=id)
    return render(request, template, {'instance':inst})


@login_required
def myrequests(request, template='goflow/myrequests.html'):
    inst_list = ProcessInstance.objects.filter(user=request.user)
    return render(request, template, {'instances':inst_list})


@login_required
def activate(request, id):
    '''
    activates and redirect to the application.
    
    parameters:
    
    id
        workitem id
    '''
    id = int(id)
    workitem = WorkItem.objects.get_safe(id=id, user=request.user)
    workitem.activate(request.user)
    return _app_response(workitem)


@login_required
def complete(request, id):
    '''
    redirect to the application.
    
    parameters:
    
    id
        workitem id
    '''
    id = int(id)
    workitem = WorkItem.objects.get_safe(id=id, user=request.user)
    return _app_response(workitem)


def _app_response(workitem):
    id = workitem.id
    activity = workitem.activity
    if not activity.process.enabled:
        return HttpResponse(gettext('process %s disabled.') % activity.process.title)
    
    if activity.kind == 'subflow':
        # subflow
        sub_workitem = workitem.start_subflow()
        return _app_response(sub_workitem)
    
    # no application: default_app
    if not activity.application:
        url = '../../../default_app'
        return HttpResponseRedirect('%s/%d/' % (url, id))
    
    if activity.kind == 'standard':
        # standard activity
        return HttpResponseRedirect(activity.application.get_app_url(workitem))
    return HttpResponse(gettext('completion page.'))




#!/usr/local/bin/python
# -*- coding: utf-8 -*-
from models import WorkItem, ProcessInstance

class ActivityState:
    blocked = 0
    inactive = 0
    active = 0
    suspended = 0
    fallout = 0
    complete = 0
    total = 0
    def __init__(self, activity):
        wis = WorkItems.objects.filter(activity=activity)
        self.total = wis.count()
        self.blocked = wis.filter(status='blocked').count()
        self.inactive = wis.filter(status='inactive').count()
        self.active = wis.filter(status='active').count()
        self.fallout = wis.filter(status='fallout').count()
        self.complete = wis.filter(status='complete').count()

class ProcessState:
    initiated = 0
    running = 0
    active = 0
    complete = 0
    terminated = 0
    suspended = 0
    total = 0
    def __init__(self, process):
        insts = ProcessInstance.objects.filter(process=process)
        self.total = insts.count()
        self.initiated = insts.filter(status='initiated').count()
        self.running = insts.filter(status='running').count()
        self.active = insts.filter(status='active').count()
        self.complete = insts.filter(status='complete').count()
        self.terminated = insts.filter(status='terminated').count()
        self.suspended = insts.filter(status='suspended').count()

class ActivityStats:
    number = 0
    time_min = None
    time_max = None
    time_mean = None
    def __init__(self, activity, user=None, year=None, month=None, day=None, datetime_interval=None):
        pass



#!/usr/local/bin/python
# -*- coding: utf-8 -*-
from django.db import models
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes.fields import GenericForeignKey

#from goflow.workflow.decorators import allow_tags
from django.utils.safestring import mark_safe
from goflow.workflow.models import Process, Activity


class Image(models.Model):
    file = models.ImageField(upload_to='images')
    info = models.CharField(max_length=100, null=True, blank=True)
    
    #@allow_tags
    def graphic(self):
        return mark_safe('<img name=image%d src=%s>' % (self.id, self.file.url))

    def __str__(self):
        return self.info or ''

class Graph(models.Model):
    name = models.CharField(max_length=100)
    parent = models.ForeignKey(
        'self', on_delete=models.CASCADE, null=True, blank=True, related_name='children')
    background = models.ForeignKey(
        'Visual', on_delete=models.CASCADE, null=True, blank=True, related_name='bg_graphes')
    def __str__(self):
        return self.name

class MetaGraph(models.Model):
    template = models.ForeignKey(Graph, on_delete=models.CASCADE)
    parent = models.ForeignKey(
        'self', on_delete=models.CASCADE, null=True, blank=True, related_name='children')
    content_type = models.ManyToManyField(ContentType)

    parent_attr = models.CharField(max_length=50, default='parent')
    children_attr = models.CharField(max_length=50, default='children')
    position_method = models.CharField(max_length=50, default='position')
    zorder_method = models.CharField(max_length=50, default='zorder')
    moveable_method = models.CharField(max_length=50, default='is_moveable')

class Visual(models.Model):
    x = models.PositiveSmallIntegerField(default=0)
    y = models.PositiveSmallIntegerField(default=0)
    w = models.PositiveSmallIntegerField(default=0)
    h = models.PositiveSmallIntegerField(default=0)
    need_update = models.BooleanField(default=True)
    visible = models.BooleanField(default=True)
    z = models.PositiveSmallIntegerField(default=0)
    image = models.ForeignKey(
        Image, on_delete=models.CASCADE, null=True, blank=True)
    
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey('content_type', 'object_id')
    
    graph = models.ForeignKey(Graph, on_delete=models.CASCADE)

    #@allow_tags
    def graphic(self):
        return mark_safe('<img src=%s>' % self.image.file.url)



from django.urls import path, re_path

from goflow.apptools.views import image_update, app_env, test_start

urlpatterns = [
    path('icon/image_update/', image_update),
    re_path(r'^application/testenv/(?P<action>create|remove)/(?P<id>.*)/$', app_env),
    path('application/teststart/<int:id>/', test_start),
]



from django.urls import path,re_path, include #django 2.*

from django.conf import settings
from django.contrib.auth.views import LoginView, LogoutView

from goflow.apptools.forms import DefaultAppStartForm
from goflow.apptools.views import start_application, default_app
from goflow.runtime.views import *
from goflow.workflow.views import process_dot, index, cron, workflow_designer, workflow_graph

urlpatterns = [
    re_path(r'^.*/logout/$', LogoutView.as_view(), name='logout'),
    re_path(r'^.*/accounts/login/$', LoginView.as_view(template_name='goflow/login.html')),
    path('apptools/', include('goflow.apptools.urls')),
    path('graph/', include('goflow.graphics.urls')),
]

urlpatterns += [
    path('', index),
    path('process/dot/<int:id>/',process_dot),
    path('cron/', cron),
    path('designer/<int:process_id>/', workflow_designer),
    path('designer/<int:process_id>/graph/', workflow_graph),
    ]

urlpatterns += [
    path('default_app/<int:id>/', default_app),
    re_path(r'^start/(?P<app_label>.*)/(?P<model_name>.*)/$', start_application),
    re_path(r'^start_proto/(?P<process_name>.*)/$', start_application,
        {'form_class':DefaultAppStartForm,
         'redirect':'../../',
         'template':'goflow/start_proto.html'}),
]

urlpatterns += [
    path('otherswork/',                 otherswork),
    path('otherswork/instancehistory/', instancehistory),
    path('myrequests/',                 myrequests),
    path('myrequests/instancehistory/', instancehistory),
    path('mywork/',                     mywork),
    path('mywork/activate/<int:id>/', activate),
    path('mywork/complete/<int:id>/', complete),
]




from django.template import Library
from django.conf import settings

register = Library()

@register.inclusion_tag('switch_user.html')
def switch_users():
    """
    Returns a menu content to switch users quickly.

    Settings required::

        settings.TEST_USERS = (('user1', 'pass1'), ...)
        settings.DEBUG = True

    FOR TESTING PURPOSE ONLY
    """
    if not settings.DEBUG: return ''
    try:
        if settings.TEST_USERS:
            pass
    except Exception:
        return ''
    
    title = 'switch user:'
    content={}
    for item in settings.TEST_USERS:
        u, p = item

        # content[u]= {' [<a href=switch/%s/%s/>%s</a>]' % (u,p,u)
        content[u] = p
    return {'title': title, 'content': content}



#!/usr/local/bin/python
# -*- coding: utf-8 -*-
from django.contrib.auth import get_user_model
from .logger import Log; log = Log('goflow.workflow.pushapps')

User = get_user_model()

def route_to_requester(workitem):
    '''Simplest possible pushapp
    '''
    return workitem.instance.user

def route_to_user(workitem, username):
    '''Route to user given a username
    '''
    return User.objects.get(username=username)

def route_to_superuser(workitem, username='admin'):
    '''Route to the superuser
    '''
    user = User.objects.get(username=username)
    if user.is_superuser:
        return user
    log.warning('this user is not a super-user:', username)
    return None

def to_current_superuser(workitem, user_pushed):
    '''Should be used in all push applications for testing purposes.
    
        (**NOT IMPLEMENTED**)
      
        usage::
        
            return to_current_superuser(workitem, user_pushed)
    '''
    return None





#!/usr/local/bin/python
# -*- coding: utf-8 -*-
from django.shortcuts import render
from django.http import HttpResponseRedirect, HttpResponse
from django.apps import apps

from django.contrib.contenttypes.models import ContentType
from django.contrib import admin

from goflow.apptools.forms import DefaultAppForm, ContentTypeForm
from goflow.apptools.models import ImageButton, Icon, Image

from goflow.workflow.models import Process, Application, Transition
from goflow.runtime.models import ProcessInstance, WorkItem

from django.db import models
from django.forms.models import modelform_factory

from django.contrib.auth.decorators import permission_required

from django.contrib.auth.decorators import login_required

from django.conf import settings
from django.utils.translation import gettext

from goflow.workflow.logger import Log; log = Log('goflow.apptools.views')

from goflow.workflow.notification import send_mail

from django.contrib import messages


@login_required
def start_application(request, app_label=None, model_name=None, process_name=None, instance_label=None,
                      template=None, template_def='goflow/start_application.html',
                      form_class=None, formset=None, redirect='home', submit_name='action',
                      ok_value='OK', cancel_value='Cancel', extra_context={}):
    '''
    generic handler for application that enters a workflow.
    
    parameters:
    
    app_label, model_name
        model linked to workflow instance (deprecated)
    process_name
        default: same name as app_label
    instance_label
        default: process_name + str(object)
    template 
        default: 'start_%s.html' % app_label
    template_def
        used if template not found - default: 'goflow/start_application.html'
    form_class
        default: django.forms.models.modelform_factory(model)
    '''
    if not process_name:
        process_name = app_label
    try:
        Process.objects.check_can_start(process_name, request.user)
    except Exception as v:
        return HttpResponse(str(v))

    # if not instance_label: instance_label = '%s-%s' % (app_label, model_name)
    if not template:
        template = 'start_%s.html' % app_label
    if not form_class:
        model = apps.get_model(app_label, model_name)
        form_class = modelform_factory(model)
        is_form_used = False
    else:
        is_form_used = True

    if not formset:
        with_formset = False
    else:
        with_formset = True

    if request.method == 'POST':
        form = form_class(request.POST, request.FILES)
        submit_value = request.POST[submit_name]
        if submit_value == cancel_value:
            return HttpResponseRedirect(redirect)

        if submit_value == ok_value and form.is_valid():
            try:
                if is_form_used:
                    ob = form.save(user=request.user, data=request.POST)
                else:
                    ob = form.save()
            except Exception as v:
                if is_form_used:
                    raise
                    log.error(
                        gettext("the save method of the form must accept parameters user and data"))
                else:
                    log.error(gettext("forme save error: %s"), str(v))

            if ob:
                if with_formset:
                    formset = formset(request.POST, request.FILES, instance=ob)
                    if formset.is_valid():
                        formset.save()

                priority = int(form.cleaned_data['priority'])
                ProcessInstance.objects.start(
                    process_name, request.user, ob, instance_label, priority=priority)

            return HttpResponseRedirect(redirect)
    else:
        form = form_class()
        # precheck
        form.pre_check(user=request.user)

        if with_formset:
            formset = formset()
    
    context = admin.site.each_context(request)   
    ex = {'form': form, 'process_name': process_name, 'formset': formset,
               'submit_name': submit_name, 'ok_value': ok_value, 'cancel_value': cancel_value}
    context.update(ex)
    
    context.update(extra_context)
    return render(request, (template, template_def), context)


@login_required
def default_app(request, id, template='goflow/default_app.html', redirect='../../', submit_name='action'):
    '''
    default application, used for prototyping workflows.
    '''
    submit_values = ('OK', 'Cancel')
    id = int(id)
    if request.method == 'POST':
        data = request.POST.copy()
        workitem = WorkItem.objects.get_safe(id, user=request.user)
        inst = workitem.instance
        ob = inst.wfobject()
        form = DefaultAppForm(data, instance=ob)
        if form.is_valid():
            # data = form.cleaned_data
            submit_value = request.POST[submit_name]
            
            workitem.instance.condition = submit_value
            
            workitem.instance.save()
            ob = form.save(workitem=workitem, submit_value=submit_value)
            # ob.comment = data['comment']
            # ob.save(workitem=workitem, submit_value=submit_value)
            
            workitem.complete(request.user)
            return HttpResponseRedirect(redirect)
    else:
        workitem = WorkItem.objects.get_safe(id, user=request.user)
        inst = workitem.instance
        ob = inst.wfobject()
        form = DefaultAppForm(instance=ob)
        # add header with activity description, submit buttons dynamically
        if workitem.activity.split_mode == 'x':
            tlist = workitem.activity.transition_inputs.all()
            if tlist.count() > 0:
                submit_values = []
                for t in tlist:
                    submit_values.append(_cond_to_button_value(t.condition))
    
    return render(request, template, {'form': form,
                                         'activity':workitem.activity,
                                         'workitem':workitem,
                                         'instance':inst,
                                         'history':inst.wfobject().history,
                                         'submit_values':submit_values, })


def _cond_to_button_value(cond):
    '''
    extract "a value" from "instance.condition=='a value'"
    used to generate buttons on default application
    '''
    import re
    s = cond.strip()
    try:
        m = re.match("instance.condition *== *(.*)", s)
        s = m.groups()[0]
        s = s.strip('"').strip("'")
    except Exception:
        pass
    return s


@login_required
def edit_model(request, id, form_class, cmp_attr=None, template=None, template_def='goflow/edit_model.html', title="",
               redirect='home', submit_name='action', ok_values=('OK',), save_value=None, cancel_value='Cancel',
               extra_context={}):
    '''
    generic handler for editing a model.
    
    parameters:

    id
        workitem id (required)
    form_class
        model form based on goflow.apptools.forms.BaseForm (required)
    cmp_attr
        edit obj.cmp_attr attribute instead of obj - default=None
    template 
        default: 'goflow/edit_%s.html' % model_lowercase
    template_def
        used if template not found - default: 'goflow/edit_model.html'
    title
        default=""
    redirect
        default='home'
    submit_name
        name for submit buttons - default='action'
    ok_values
        submit buttons values - default=('OK',)
    save_value
        save button value - default='Save'
    cancel_value
        cancel button value - default='Cancel'
    extra_context
        default={}
    '''
    if not template: template = 'goflow/edit_%s.html' % form_class._meta.model._meta.object_name.lower()
    model_class = form_class._meta.model
    workitem = WorkItem.objects.get_safe(int(id), user=request.user)
    instance = workitem.instance
    activity = workitem.activity
    
    obj = instance.wfobject()
    obj_context = obj
    # objet composite intermédiaire
    if cmp_attr:
        obj = getattr(obj, cmp_attr)
    
    template = override_app_params(activity, 'template', template)
    redirect = override_app_params(activity, 'redirect', redirect)
    submit_name = override_app_params(activity, 'submit_name', submit_name)
    ok_values = override_app_params(activity, 'ok_values', ok_values)
    cancel_value = override_app_params(activity, 'cancel_value', cancel_value)

    if request.method == 'POST':
        form = form_class(request.POST, instance=obj)
        submit_value = request.POST[submit_name]
        if submit_value == cancel_value:
            return HttpResponseRedirect(redirect)
        
        if form.is_valid():
            if (submit_value == save_value):
                # just save
                # ob = form.save()
                try:
                    ob = form.save(workitem=workitem, submit_value=submit_value)
                except Exception as v:
                    raise Exception(str(v))
                return HttpResponseRedirect(redirect)
            
            if submit_value in ok_values:
                # save and complete activity
                # ob = form.save()
                try:
                    ob = form.save(workitem=workitem, submit_value=submit_value)
                except Exception as v:
                    raise Exception(str(v))
                instance.condition = submit_value
                instance.save()
                workitem.complete(request.user)
                return HttpResponseRedirect(redirect)
    else:
        form = form_class(instance=obj)
        # precheck
        form.pre_check(obj_context, user=request.user)
    
    context = {  'form': form, 'object':obj, 'object_context':obj_context,
                 'instance':instance, 'workitem':workitem,
                 'submit_name':submit_name, 'ok_values':ok_values,
                 'save_value':save_value, 'cancel_value':cancel_value,
                 'title':title}
    context.update(extra_context)
    print((template, template_def))
    print(context)
    return render(request, (template, template_def), context)


@login_required
def view_application(request, id, template='goflow/view_application.html', redirect='home', title="",
               submit_name='action', ok_values=('OK',), cancel_value='Cancel',
               extra_context={}):
    '''
    generic handler for a view application.
    
    useful for a simple view or a complex object edition.

    parameters:
    
    id
        workitem id (required)
    template 
        default: 'goflow/view_application.html'
    redirect
        default='home'
    title
        default=""
    submit_name
        name for submit buttons - default='action'
    ok_values
        submit buttons values - default=('OK',)
    cancel_value
        cancel button value - default='Cancel'
    extra_context
        default={}
    '''
    workitem = WorkItem.objects.get_safe(int(id), user=request.user)
    instance = workitem.instance
    activity = workitem.activity
    
    obj = instance.wfobject()
    
    template = override_app_params(activity, 'template', template)
    redirect = override_app_params(activity, 'redirect', redirect)
    submit_name = override_app_params(activity, 'submit_name', submit_name)
    ok_values = override_app_params(activity, 'ok_values', ok_values)
    cancel_value = override_app_params(activity, 'cancel_value', cancel_value)

    if request.method == 'POST':
        submit_value = request.POST[submit_name]
        if submit_value == cancel_value:
            return HttpResponseRedirect(redirect)
        
        if submit_value in ok_values:
            instance.condition = submit_value
            instance.save()
            workitem.complete(request.user)
            return HttpResponseRedirect(redirect)
        
    context = {  'object':obj, 'instance':instance, 'workitem':workitem,
                 'submit_name':submit_name,
                 'ok_values':ok_values, 'cancel_value':cancel_value,
                 'title':title}
    context.update(extra_context)
    return render(request, template, context)


@login_required
def choice_application(request, id, template='goflow/view_application_image.html', redirect='home', title="Choice",
               submit_name='image', cancel_action='cancel', extra_context={}):
    '''
    a view to make a choice within image buttons.
    
    - actions are generated from instances conditions of outer transitions
    - the activity split_mode must be xor
    - actions are rendered with images
    - actions are mapped to images with ImageButton instances

    parameters:
    
    id
        workitem id (required)
    template 
        default: 'goflow/view_application_image.html'
    redirect
        default='home'
    title
        default='Choice'
    submit_name
        name for submit buttons - default='image'
    cancel_value
        cancel button value - default='Cancel'
    extra_context
        default={}
    '''
    workitem = WorkItem.objects.get_safe(int(id), user=request.user)
    activity = workitem.activity
    if activity.split_mode != 'xor':
        raise Exception('choice_application: split_mode xor required')
    list_trans = Transition.objects.filter(input=activity)
    ok_values = []
    for t in list_trans:
        if ImageButton.objects.filter(action=t.condition).count() == 0:
            raise Exception('no ImageButton for action [%s]' % t.condition)
        ok_values.append(t.condition)
        
    return view_application(request, id, template, redirect, title,
               submit_name, ok_values, cancel_action, extra_context)


def sendmail(workitem, subject='goflow.apptools sendmail message', template='goflow/app_sendmail.txt'):
    '''send a mail notification to the workitem user.
    
    parameters:
    
    subject
        default='goflow.apptools sendmail message'
    template
        default='goflow/app_sendmail.txt'
    '''
    send_mail(workitems=(workitem,), user=workitem.user, subject=subject, template=template)


def override_app_params(activity, name, value):
    '''
    usage: param = _override_app_params(activity, 'param', param)
    '''
    try:
        if not activity.app_param:
            return value
        dicparams = eval(activity.app_param)
        # if dicparams.has_key(name):
        if name in dicparams.keys():
            return dicparams[name]
    except Exception as v:
        log.error('_override_app_params %s %s - %s', activity, name, v)
    return value


@login_required
def app_env(request, action, id, template=None):
    """
    creates/removes unit test environment for applications.
    
    a process named "test_[app]" with one activity
    a group with appropriate permission
    """
    app = Application.objects.get(id=int(id))
    rep = 'Nothing done.'
    if action == 'create':
        app.create_test_env(user=request.user)
        rep = 'test env created for app %s' % app.url
    if action == 'remove':
        app.remove_test_env()
        rep = 'test env removed for app %s' % app.url
    
    rep += '<hr><p><b><a href=../../../>return</a></b>'
    return HttpResponse(rep)


@login_required
def test_start(request, id, template='goflow/test_start.html'):
    """
    starts test instances.
    
    for a given application, with its unit test environment, the user
    choose a content-type then generates unit test process instances
    by cloning existing content-type objects (**Work In Progress**).
    """
    app = Application.objects.get(id=int(id))
    context = {}
    if request.method == 'POST':
        submit_value = request.POST['action']
        if submit_value == 'Create':
            ctype = ContentType.objects.get(id=int(request.POST['ctype']))
            model = ctype.model_class()
            for inst in model.objects.all():
                # just objects without link to a workflow instance
                if ProcessInstance.objects.filter(
                    content_type__pk=ctype.id,
                    object_id=inst.id
                ).count() > 0:
                    continue
                inst.id = None
                inst.save()
                # TODO: convert this to method
                ProcessInstance.objects.start(
                # start_instance(
                            process_name='test_%s' % app.url,
                            user=request.user, item=inst,
                            title="%s test instance for app %s" % (
                                ctype.name, app.url
                            ))
            # request.user.message_set.create(message='test instances created')
            messages.add_message(request, messages.INFO,
                                 "test instances created")
        return HttpResponseRedirect('../..')
    form = ContentTypeForm()
    context['form'] = form
    return render(request, template, context)


@login_required
def image_update(request):
    '''
    Import Image instances as Icon instances.
    
    GoFlow can use local images as well as http distant images.
    For a genericity reason, local images (Image) are obtained as
    Icon images by their url; this view is useful to wrap all local
    images in a row. 
    '''
    rep = gettext('<h1>Update Icons from Images</h1>')
    for im in Image.objects.all():
        if Icon.objects.filter(url__endswith=str(im.url)).count() == 0:
            ic, created = Icon.objects.get_or_create(category='local-' + im.category, url=im.url())
            if created: rep += gettext('<br> %s added ') % im.url()
    rep += gettext('<hr><p><b><a href=../>return</a></b>')
    return HttpResponse(rep)
    



from django.contrib import admin
from django.db.models import *

from django.contrib.auth import get_user_model
from django.contrib.auth.admin import UserAdmin

from goflow.workflow.models import Transition, Process, Application, PushApplication, Activity, UserProfile

User = get_user_model()


class TransitionInline(admin.StackedInline):
    model = Transition
    fieldsets = (
              (None, {'fields':(('input', 'output'), 'condition', 'precondition')}),
              )


class ProcessAdmin(admin.ModelAdmin):
    list_display = ('title', 'enabled', 'summary', 'priority')
    inlines = [
                   TransitionInline,
               ]


admin.site.register(Process, ProcessAdmin)


class ApplicationAdmin(admin.ModelAdmin):
    save_as = True
    list_display = ('url', 'documentation', 'test')


admin.site.register(Application, ApplicationAdmin)


class PushApplicationAdmin(admin.ModelAdmin):
    save_as = True
    list_display = ('url', 'documentation', 'test')


admin.site.register(PushApplication, PushApplicationAdmin)


class ActivityAdmin(admin.ModelAdmin):
    save_as = True
    list_display = ('title', 'description', 'kind', 'application',
                    'join_mode', 'split_mode', 'autostart', 'autofinish', 'process')
    list_filter = ('process', 'kind')
    fieldsets = (
              (None, {'fields':(('kind', 'subflow'), ('title', 'process'), 'description')}),
              ('Push application', {'fields':(('push_application', 'pushapp_param'),)}),
              ('Application', {'fields':(('application', 'app_param'),)}),
              ('I/O modes', {'fields':(('join_mode', 'split_mode'),)}),
              ('Execution modes', {'fields':(('autostart', 'autofinish'),)}),
              ('Permission', {'fields':('roles',)}),
              )


admin.site.register(Activity, ActivityAdmin)


class TransitionAdmin(admin.ModelAdmin):
    save_as = True
    list_display = ('__str__', 'input', 'output', 'condition', 'description', 'process')
    list_filter = ('process',)
    fieldsets = (
              (None, {'fields':(
                                ('name', 'description'),
                                'process',
                                ('input', 'output'),
                                'condition', 'precondition'
                                )
                     }),
              )


admin.site.register(Transition, TransitionAdmin)


class UserProfileAdmin(admin.ModelAdmin):
    list_display = ('user', 'web_host', 'notified', 'last_notif', 'nb_wi_notif', 'notif_delay')
    list_filter = ('web_host', 'notified')


admin.site.register(UserProfile, UserProfileAdmin)

admin.site.unregister(User)


class UserProfileInline(admin.StackedInline):
    model = UserProfile
    max_num = 1


class GoFlowUserAdmin(UserAdmin):
    inlines = [UserProfileInline]


admin.site.register(User, GoFlowUserAdmin)



from django.contrib import admin

from goflow.apptools.models import Image, Icon, ImageButton
from goflow.workflow.models import Transition
from goflow.workflow.admin import TransitionAdmin as TransitionAdminOld

class ImageAdmin(admin.ModelAdmin):
    list_display = ('category', 'graphic', 'file', 'url')
    list_filter = ('category',)
admin.site.register(Image, ImageAdmin)

class IconAdmin(admin.ModelAdmin):
    list_display = ('category', 'graphic', 'url')
    list_filter = ('category',)
admin.site.register(Icon, IconAdmin)

class ImageButtonAdmin(admin.ModelAdmin):
    raw_id_fields = ('icon',)
    list_display = ('action', 'label', 'graphic')
admin.site.register(ImageButton, ImageButtonAdmin)



from pathlib import Path

from setuptools import setup


BASE_DIR = Path(__file__).resolve().parent
README_PATH = BASE_DIR / 'README.md'

setup(
    name='django-goflow2',
    version='1.0',
    description='Fork from Eric Simorre\'s django-goflow, updated for modern Python and Django 4/5/6',
    long_description=README_PATH.read_text(encoding='utf-8'),
    long_description_content_type='text/markdown',
    author='Mike Li',
    author_email='mikewolfli@163.com',
    url='https://github.com/mikewolfli/django-goflow',
    project_urls={
        'Source': 'https://github.com/mikewolfli/django-goflow',
        'Tracker': 'https://github.com/mikewolfli/django-goflow/issues',
    },
    keywords='django workflow bpm goflow',
    packages=['goflow', 'goflow.apptools', 'goflow.runtime', 'goflow.graphics', 'goflow.graphics2', 'goflow.workflow', 'goflow.apptools.templatetags', 'goflow.runtime.templatetags', 'goflow.workflow.templatetags', 'goflow.graphics2.templatetags'],
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Framework :: Django :: 4.2',
        'Framework :: Django :: 5.0',
        'Framework :: Django :: 6.0',
    ],
    python_requires='>=3.9',
    install_requires=[
        'Django>=4.2,<7.0',
        'django-ninja>=1.5.3',
        'django-ninja-simple-jwt>=0.7.1',
        'Pillow>=10.0.0',
    ],
    include_package_data=True,
    zip_safe=False,
)



from django.contrib import admin

from goflow.runtime.models import ProcessInstance, WorkItem, Event


class ProcessInstanceAdmin(admin.ModelAdmin):
    date_hierarchy = 'creationTime'
    list_display = ('title', 'process', 'user', 'creationTime', 'status', 'workitems_list')
    list_filter = ('process', 'status', 'user')
    fieldsets = (
              (None, {'fields':(
                                'title', 'process', 'user',
                                ('status', 'old_status'),
                                'condition',
                                ('object_id', 'content_type'))
                     }),
              )
admin.site.register(ProcessInstance, ProcessInstanceAdmin)


class WorkItemAdmin(admin.ModelAdmin):
    date_hierarchy = 'date'
    list_display = ('date', 'user', 'instance', 'activity', 'status', 'events_list')
    list_filter = ('status', 'user', 'activity',)
    fieldsets = (
              (None, {'fields':(
                                ('instance', 'activity'),
                                'user',
                                'workitem_from',
                                ('status', 'blocked', 'priority'),
                                'push_roles', 'pull_roles')
                     }),
              )
admin.site.register(WorkItem, WorkItemAdmin)


class EventAdmin(admin.ModelAdmin):
    date_hierarchy = 'date'
    list_display = ('date', 'name', 'workitem')
admin.site.register(Event, EventAdmin)



from django.urls import path, re_path

from goflow.apptools import forms
from goflow.apptools.views import *

urlpatterns = [
    re_path(r'^start/(?P<app_label>.*)/(?P<model_name>.*)/$', start_application),
    re_path(r'^start_proto/(?P<process_name>.*)/$', start_application, {'form_class':forms.DefaultAppStartForm, 'template':'goflow/start_proto.html'}),
    re_path(r'^view_application/(?P<id>\d+)/$', view_application),
    re_path(r'^choice_application/(?P<id>\d+)/$', choice_application),
    path('sendmail/', sendmail),
]



#!/usr/local/bin/python
# -*- coding: utf-8 -*-
from django.db import models
from goflow.workflow.models import Transition
#from goflow.workflow.decorators import allow_tags
from django.utils.safestring import mark_safe
from django.conf import settings
from django.utils.translation import gettext_lazy as _, gettext


class DefaultAppModel(models.Model):
    """Default implementation object class  for process instances.
    
    When a process instance starts, the instance has to carry an
    implementation object that contains the application data. The
    specifications for the implementation class is:
    
    (nothing: now managed by generic relation)
    
    This model is used in process simulations: you don't have to define
    application in activities for this; the DefaultAppModel is used
    to keep workflow history for displaying to users.
    """
    history = models.TextField(_('history'), editable=False, null=True, blank=True)
    comment = models.TextField(_('comment'), null=True, blank=True)
    
    def __str__(self):
        return gettext('simulation model %s') % str(self.id)
    class Admin:
        list_display = ('__str__',)
    class Meta:
        verbose_name=_('Simulation object')

class Image(models.Model):
    '''
    An image stored in the database
    '''
    category = models.CharField(_('category'), max_length=20, null=True, blank=True)
    file = models.ImageField(_('file'), upload_to='images')
    
    def url(self):
        return "%s%s" % (settings.MEDIA_URL, self.file)
    
    #@allow_tags
    def graphic(self):
        '''
        generates an *img* html tag for html rendering
        '''
        return mark_safe('<img name="image%d" src="%s">' % (self.pk, self.url()))
    
    #@allow_tags
    def graphic_input(self):
        '''
        generates an *input* html tag with type=image for html rendering
        '''
        return mark_safe('<input type=image name=icon src=%s>' % self.url())
    
    def __str__(self):
        return str(self.file)

    class Meta:
        verbose_name=_('Image')

class Icon(models.Model):
    '''
    An image accessible by an url.

    Tip: all of the "Image" objects can be imported as "Icon" from the
    admin panel.
    '''
    category = models.CharField(_('catagory'), max_length=20, null=True, blank=True)
    # url = models.URLField(verify_exists=False)
    url = models.URLField()
    
    #@allow_tags
    def graphic(self):
        '''
        generates an *img* html tag for html rendering
        '''
        return mark_safe('<img name="image%d" src="%s">' % (self.pk, self.url))
    
    #@allow_tags
    def graphic_input(self):
        '''
        generates an *input* html tag with type=image for html rendering
        '''
        return mark_safe('<input type=image name=icon src="%s">' % self.url)
    
    def __str__(self):
        return self.url

    class Meta:
        verbose_name=_('Icon')

class ImageButton(models.Model):
    '''
    Mapping object between an "action" and an "Icon".
    
    ImageButton objects have also a textual field: label.
    '''
    action = models.SlugField(primary_key=True,verbose_name=_('action'))
    label = models.CharField(_('label'), max_length=100)
    icon = models.ForeignKey(Icon, on_delete=models.CASCADE, verbose_name=_('icon'))
    
    #@allow_tags
    def graphic(self):
        '''
        generates an *img* html tag for html rendering
        '''
        return mark_safe('<img name="image-%s" src="%s">' % (self.pk, self.icon.url))
    
    #@allow_tags
    def graphic_input(self):
        '''
        generates an *input* html tag with type=image for html rendering
        '''
        return mark_safe('<input type=image name=image src="%s" value="%s" title="%s">' % (self.icon.url, self.pk, self.label))
    
    def __str__(self):
        return self.label

    class Meta:
        verbose_name=_("Image Button")
